
#include <boost/shared_ptr.hpp>
#include <iostream>


using boost::shared_ptr;
using std::cout;
using std::endl;


template<typename Key>
struct node
{
	Key key;
	
	shared_ptr<node<Key> > left;
	
	shared_ptr<node<Key> > right;
	
	shared_ptr<node<Key> > parent;
	
	short balance;
	
	node() : key(0), left(0), right(0), parent(0), balance(0)
	{
		cout << "node::node():" << endl;
	}
	
	node(Key a_key) : key(a_key), balance(0)
	{
		cout << "node::node(): a_key=" << a_key << endl;
	}
	
	~node()
	{
		cout << "node::~node(): key=" << key << ", right.use_count()=" << right.use_count() << endl;
	}
};


template<typename Key>
class avl_tree
{
public:

	explicit avl_tree();
	
	~avl_tree();

	shared_ptr<node<Key> > find(Key key);

	void insert(Key key);

	void remove(Key key);

private:

	shared_ptr<node<Key> > predecessor(shared_ptr<node<Key> > a_node);

	shared_ptr<node<Key> > successor(shared_ptr<node<Key> > a_node);

	void rotate_left(shared_ptr<node<Key> > a_node);

	void rotate_right(shared_ptr<node<Key> > a_node);

	void case_l1(shared_ptr<node<Key> > x);

	void case_l2(shared_ptr<node<Key> > x);

	void case_l3(shared_ptr<node<Key> > x);

	void case_l4(shared_ptr<node<Key> > x);

	void case_l5(shared_ptr<node<Key> > x);

	void case_r1(shared_ptr<node<Key> > x);

	void case_r2(shared_ptr<node<Key> > x);

	void case_r3(shared_ptr<node<Key> > x);

	void case_r4(shared_ptr<node<Key> > x);

	void case_r5(shared_ptr<node<Key> > x);

	bool is_left_descendant(shared_ptr<node<Key> > x, shared_ptr<node<Key> > y);

	void destroy_subtree(shared_ptr<node<Key> > root);

	avl_tree(const avl_tree&);

	void operator=(const avl_tree&);

	shared_ptr<node<Key> > _root;
};


